/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <libxml/tree.h>
#include <glib.h>
#include "nntpgrab_utils.h"

static char *xml_parser_errmsg = NULL;

static void
xml_parser_error_func (void *ctx, const char *msg, ...)
{
    va_list args;
    char *orig = xml_parser_errmsg;
    char *tmp;

    va_start(args, msg);

    tmp = g_strdup_vprintf(msg, args);
    if (orig) {
        xml_parser_errmsg = g_strdup_printf("%s %s", orig, tmp);
        g_free(orig);
        g_free(tmp);
    } else {
        xml_parser_errmsg = tmp;
    }

    va_end(args);
}

static void
parse_groups(xmlDocPtr doc, xmlNodePtr cur, NZBFile *file)
{
    while(cur != NULL) {
        if (!xmlStrcmp(cur->name, (const xmlChar *) "group")) {
            xmlChar *group = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            file->groups = ng_list_append(file->groups, g_strdup((const gchar *) group));
            xmlFree(group);
        }

        cur = cur->next;
    }
}

static void
parse_segments(xmlDocPtr doc, xmlNodePtr cur, NZBFile *file)
{
    while(cur != NULL) {
        if (!xmlStrcmp(cur->name, (const xmlChar *) "segment")) {
            NNTPGrabPart *segment = g_slice_new0(NNTPGrabPart);
            xmlChar *value;
            char *message_id;

            value = xmlGetProp(cur, (const xmlChar *) "bytes");
            segment->size = atol((const char *) value);
            file->filesize += segment->size;
            xmlFree(value);

            value = xmlGetProp(cur, (const xmlChar *) "number");
            segment->part_num = atoi((const char *) value);
            xmlFree(value);

            // NZB files contain a message-id without the brackets. Add them manually
            value = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            message_id = g_strdup_printf("<%s>", (const gchar *) value);
            strncpy((char*) segment->message_id, message_id, (size_t) sizeof(segment->message_id) - 1);
            g_free(message_id);
            xmlFree(value);

            file->segments = ng_list_append(file->segments, segment);
        }

        cur = cur->next;
    }
}

static void
parse_file(xmlDocPtr doc, xmlNodePtr cur, NZBFile *file)
{
    while(cur != NULL) {
        if (!xmlStrcmp(cur->name, (const xmlChar *) "groups")) {
            parse_groups(doc, cur->xmlChildrenNode, file);
        }

        if (!xmlStrcmp(cur->name, (const xmlChar *) "segments")) {
            parse_segments(doc, cur->xmlChildrenNode, file);
        }

        cur = cur->next;
    }
}

NZB *
nntpgrab_utils_parse_nzb_file(const char *contents, char **errmsg)
{
    xmlDocPtr    doc;
    xmlNodePtr   cur;
    NZB         *nzb = NULL;

    xmlSetGenericErrorFunc(NULL, xml_parser_error_func);

    if (xml_parser_errmsg) {
        g_free(xml_parser_errmsg);
        xml_parser_errmsg = NULL;
    }

    doc = xmlParseMemory(contents, strlen(contents));
    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        if (errmsg) {
            *errmsg = g_strdup(xml_parser_errmsg);
        }

        g_free(xml_parser_errmsg);
        xml_parser_errmsg = NULL;

        xmlFreeDoc(doc);
        return NULL;
    }

    if (xmlStrcmp(cur->name, (const xmlChar *) "nzb")) {
        if (errmsg) {
            *errmsg = g_strdup("parse_nzb_file: document of the wrong type, root node != nzb");
        }

        xmlFreeDoc(doc);

        return FALSE;
    }

    cur = cur->xmlChildrenNode;
    if (!cur) {
        if (errmsg) {
            *errmsg = g_strdup("parse_nzb_file: no children of root-node!");
        }

        xmlFreeDoc(doc);

        return FALSE;
    }

    nzb = g_slice_new0(NZB);

    do {
        if (!xmlStrcmp(cur->name, (const xmlChar *) "file")) {
            NZBFile *file = g_slice_new0(NZBFile);
            xmlChar *value;

            value = xmlGetProp(cur, (const xmlChar *) "subject");
            strncpy((char *) file->subject, (const char *) value, (size_t) sizeof(file->subject));
            xmlFree(value);

            value = xmlGetProp(cur, (const xmlChar *) "poster");
            strncpy((char *) file->poster, (const char *) value, (size_t) sizeof(file->poster));
            xmlFree(value);

            value = xmlGetProp(cur, (const xmlChar *) "date");
            file->stamp = atoi((const char *) value);
            xmlFree(value);

            parse_file(doc, cur->xmlChildrenNode, file);

            nzb->files = ng_list_append(nzb->files, file);
        }
    } while ((cur=cur->next));

    xmlFreeDoc(doc);

    return nzb;
}

void
nntpgrab_utils_nzb_file_free(NZB *nzbfile)
{
    NGList *list;

    list = nzbfile->files;
    while (list) {
        NZBFile *file = (NZBFile *) list->data;
        NGList *list2;

        list2 = file->groups;
        while (list2) {
            g_free(list2->data);
            list2 = ng_list_next(list2);
        }

        ng_list_free(file->groups);

        list2 = file->segments;
        while (list2) {
            g_slice_free(NNTPGrabPart, list2->data);
            list2 = ng_list_next(list2);
        }

        ng_list_free(file->segments);

        g_slice_free(NZBFile, file);

        list = ng_list_next(list);
    }

    g_slice_free(NZB, nzbfile);
}
