/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <glib/gi18n.h>
#include "gui.h"
#include "nntpgrab_utils.h"

static gboolean scroll_to_cell = FALSE;
static gboolean initialized = FALSE;

static char *
retrieve_path(GtkTreeModel *model, GtkTreeIter tree_iter)
{
    GtkTreeIter iter = tree_iter;
    GtkTreeIter child;
    GtkTreeIter iter_tmp;
    char *path = NULL;
    char *dir_separator = "/";

    // Find out of the server is a UNIX or a Win32 environment
    if (gtk_tree_model_get_iter_first(model, &iter_tmp)) {
        char *root = NULL;

        gtk_tree_model_get(model, &iter_tmp, 0, &root, -1);

        if (root && !strcmp(root, "/")) {
            dir_separator = "/";
        } else {
            dir_separator = "\\";
        }

        g_free(root);
    }

    do {
        char *dirname = NULL;
        char *tmp;

        child = iter;

        gtk_tree_model_get(model, &child, 0, &dirname, -1);

        if (path) {
            if (!strcmp(dirname, "/")) {
                // Don't use the dir separator for the root
                tmp = g_strdup_printf("%s%s", dirname, path);
                g_free(path);
                path = tmp;
            } else {
                tmp = g_strdup_printf("%s%s%s", dirname, dir_separator, path);      // We walk the path from right to left
                g_free(path);
                path = tmp;
            }
        } else {
            path = g_strdup(dirname);
        }

        g_free(dirname);
    } while (gtk_tree_model_iter_parent(model, &iter, &child));

    return path;
}

static void
on_row_expanded (GtkTreeView *tree_view, GtkTreeIter *tree_iter, GtkTreePath *tree_path, gpointer user_data)
{
    char *path = NULL;
    GtkTreeModel *model = gtk_tree_view_get_model(tree_view);
    NGList *folders = NULL;
    GtkTreeIter iter;

    memcpy(&iter, tree_iter, sizeof(GtkTreeIter));

    path = retrieve_path(model, iter);

    if (nntpgrab_config_get_folder_listing(glue, path, &folders)) {
        NGList *list = folders;
        GtkTreeIter iter;

        while (list) {
            NNTPGrabFolder *folder = (NNTPGrabFolder*) list->data;
            gboolean found = FALSE;
            gboolean create_new = TRUE;

            // Do we already have a child with this name ?
            if (gtk_tree_model_iter_children(model, &iter, tree_iter)) {
                char *folder2 = NULL;

                do {
                    gtk_tree_model_get(model, &iter, 0, &folder2, -1);

                    if (!folder2) {
                        create_new = FALSE;
                        break;
                    }

                    if (folder2 && !strcmp(folder->folder, folder2)) {
                        found = TRUE;
                    }

                    g_free(folder2);
                } while (gtk_tree_model_iter_next(model, &iter));
            }

            if (found) {
                list = ng_list_next(list);
                continue;
            }

            if (create_new) {
                gtk_tree_store_append(GTK_TREE_STORE(model), &iter, tree_iter);
            }

            gtk_tree_store_set(GTK_TREE_STORE(model), &iter, 0, folder->folder, -1);

            if (folder->has_subfolders && !gtk_tree_model_iter_has_child(model, &iter)) {
                GtkTreeIter iter_ignore;
                gtk_tree_store_append(GTK_TREE_STORE(model), &iter_ignore, &iter);
            }

            list = ng_list_next(list);
        }

        nntpgrab_config_free_folder_listing(glue, folders);
    }

    g_free(path);
}

static void
update_folder_listing(GtkTreeView *tree, GtkTreeStore *store, GtkTreeSelection *selection, const char *directory)
{
    char **values;
    char *dir_separator;
    int i;
    GtkTreeIter parent;
    GtkTreePath *tree_path;
    char *path = NULL;

    gtk_tree_store_clear(store);

    if (strlen(directory) >= 3 && directory[1] == ':' && directory[2] == '\\') {
        dir_separator = "\\";
    } else {
        dir_separator = "/";
    }

    memset(&parent, 0, sizeof(GtkTreeIter));

    values = g_strsplit(directory, dir_separator, 0);
    i = 0;
    while (values[i]) {
        GtkTreeIter iter;
        NGList *folders;
        NGList *list;
        GtkTreeIter temp;
        gboolean already_has_child;

        // Ignore empty paths
        if (strlen(values[i]) == 0 && i > 0) {
            i++;
            continue;
        }

        // build the path (necessary for the nntpgrab_config_get_folder_listing() call)
        if (path) {
            char *tmp;
            if (dir_separator[0] == '/' && i == 1) {
                tmp = g_strdup_printf("%s%s", path, values[i]);
            } else {
                tmp = g_strdup_printf("%s%s%s", path, dir_separator, values[i]);
            }
            g_free(path);
            path = tmp;
        } else {
            if (strlen(values[i]) == 0) {
                path = g_strdup("/");
            } else {
                path = g_strdup(values[i]);
            }
        }

        // build the tree
        if (i == 0) {
            // Find out all the available drives on the NNTPGrab server
            // and add these drives to the list
            NGList *list;

            g_return_if_fail(nntpgrab_config_get_folder_listing(glue, NULL, &folders));

            list = folders;
            while (list) {
                NNTPGrabFolder *folder = (NNTPGrabFolder*) list->data;

                gtk_tree_store_append(store, &iter, NULL);
                gtk_tree_store_set(store, &iter, 0, folder->folder, -1);

                if (!strcmp(folder->folder, path)) {
                    parent = iter;
                }

                if (folder->has_subfolders) {
                    GtkTreeIter iter_ignore;
                    gtk_tree_store_append(store, &iter_ignore, &iter);
                }

                list = ng_list_next(list);
            }

            nntpgrab_config_free_folder_listing(glue, folders);
        }

        g_return_if_fail(nntpgrab_config_get_folder_listing(glue, path, &folders));

        list = folders;
        temp = parent;

        // Is there already an child created?
        if (gtk_tree_model_iter_children(GTK_TREE_MODEL(store), &iter, &parent)) {
            already_has_child = TRUE;
            if (!strcmp(path, directory)) {
                gtk_tree_store_remove(store, &iter);
                already_has_child = FALSE;
            }
        } else {
            already_has_child = FALSE;
        }

        while (list) {
            NNTPGrabFolder *folder = (NNTPGrabFolder*) list->data;

            if (!already_has_child) {
                gtk_tree_store_append(store, &iter, &parent);
            } else {
                already_has_child = FALSE;
            }

            gtk_tree_store_set(store, &iter, 0, folder->folder, -1);

            if (folder->has_subfolders) {
                GtkTreeIter iter_ignore;
                gtk_tree_store_append(store, &iter_ignore, &iter);
            }

            if (values[i + 1] && !strcmp(folder->folder, values[i + 1])) {
                temp = iter;
            }

            list = ng_list_next(list);
        }

        parent = temp;

        nntpgrab_config_free_folder_listing(glue, folders);

        i++;
    }

    tree_path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &parent);
    gtk_tree_view_expand_to_path(tree, tree_path);
    gtk_tree_selection_select_iter(selection, &parent);
    gtk_tree_view_scroll_to_cell(tree, tree_path, NULL, TRUE, 0.5, 0.0);
    gtk_tree_path_free(tree_path);

    g_strfreev(values);
    if (path) {
        g_free(path);
    }
}

enum {
    TREE_DOWNLOAD_DIRECTORY,
    TREE_TEMP_DIRECTORY
};

static void
on_selection_changed (GtkTreeSelection *selection, gpointer user_data)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    char *path;
    ConfigOpts opts;

    if (!gtk_tree_selection_get_selected(selection, &model, &iter)) {
        return;
    }

    scroll_to_cell = FALSE;

    path = retrieve_path(model, iter);

    opts = nntpgrab_config_get_opts(glue);
    switch (GPOINTER_TO_INT(user_data)) {
        case TREE_DOWNLOAD_DIRECTORY:
            if (!strcmp(opts.download_directory, path)) {
                // Nothing changed, ignore
                g_free(path);
                return;
            }

            memset(opts.download_directory, 0, sizeof(opts.download_directory));
            strncpy(opts.download_directory, path, sizeof(opts.download_directory) - 1);
            break;

        case TREE_TEMP_DIRECTORY:
            if (!strcmp(opts.temp_directory, path)) {
                // Nothing changed, ignore
                g_free(path);
                return;
            }

            memset(opts.temp_directory, 0, sizeof(opts.temp_directory));
            strncpy(opts.temp_directory, path, sizeof(opts.temp_directory) - 1);
            break;

        default:
            g_print(__FILE__ ":%i Invalid user_data: %i\n", __LINE__, GPOINTER_TO_INT(user_data));
            break;
    };

    scroll_to_cell = FALSE;
    nntpgrab_config_set_opts(glue, opts);

    g_free(path);
}

static gboolean
config_changed_phase2(gpointer data)
{
    GtkWidget *treeConfigDownloadDirectory;
    GtkWidget *treeConfigTempDirectory;
    GtkTreeStore *storeDownload;
    GtkTreeStore *storeTemp;
    GtkTreeSelection *selection;
    ConfigOpts opts;
    static ConfigOpts prev_opts;
    static gboolean beenhere = FALSE;
    GtkWidget *window = GTK_WIDGET(data);

    if (!beenhere) {
        beenhere = TRUE;
        memset(&prev_opts, 0, sizeof(ConfigOpts));
    }

    opts = nntpgrab_config_get_opts(glue);

    // Did we cause the signal to be emitted ?
    if (!scroll_to_cell) {
        scroll_to_cell = TRUE;
        prev_opts = opts;
        return FALSE;
    }

    if (strcmp(opts.download_directory, prev_opts.download_directory)) {
        treeConfigDownloadDirectory = nntpgrab_gui_base_get_widget("treeConfigDownloadDirectory");

        storeDownload = GTK_TREE_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(treeConfigDownloadDirectory)));
        selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeConfigDownloadDirectory));
        g_signal_handlers_block_by_func (G_OBJECT (treeConfigDownloadDirectory), G_CALLBACK (on_row_expanded), NULL);
        g_signal_handlers_block_by_func (G_OBJECT (selection), G_CALLBACK (on_selection_changed), GINT_TO_POINTER(TREE_DOWNLOAD_DIRECTORY));
        update_folder_listing(GTK_TREE_VIEW(treeConfigDownloadDirectory), storeDownload, selection, opts.download_directory);
        g_signal_handlers_unblock_by_func (G_OBJECT (selection), G_CALLBACK (on_selection_changed), GINT_TO_POINTER(TREE_DOWNLOAD_DIRECTORY));
        g_signal_handlers_unblock_by_func (G_OBJECT (treeConfigDownloadDirectory), G_CALLBACK (on_row_expanded), NULL);
    }

    if (strcmp(opts.temp_directory, prev_opts.temp_directory)) {
        treeConfigTempDirectory = nntpgrab_gui_base_get_widget("treeConfigTempDirectory");

        storeTemp = GTK_TREE_STORE(gtk_tree_view_get_model(GTK_TREE_VIEW(treeConfigTempDirectory)));
        selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeConfigTempDirectory));
        g_signal_handlers_block_by_func (G_OBJECT (treeConfigTempDirectory), G_CALLBACK (on_row_expanded), NULL);
        g_signal_handlers_block_by_func (G_OBJECT (selection), G_CALLBACK (on_selection_changed), GINT_TO_POINTER(TREE_TEMP_DIRECTORY));
        update_folder_listing(GTK_TREE_VIEW(treeConfigTempDirectory), storeTemp, selection, opts.temp_directory);
        g_signal_handlers_unblock_by_func (G_OBJECT (selection), G_CALLBACK (on_selection_changed), GINT_TO_POINTER(TREE_TEMP_DIRECTORY));
        g_signal_handlers_unblock_by_func (G_OBJECT (treeConfigTempDirectory), G_CALLBACK (on_row_expanded), NULL);
    }

    gtk_widget_hide(window);
    gtk_widget_destroy(window);

    prev_opts = opts;

    return FALSE;
}

static void
config_changed(NNTPGrabGlue *glue, gpointer data)
{
    GtkWidget *window = NULL;
    GtkWidget *label;
    char *markup;

    if (!initialized) {
        return;
    }

    if (scroll_to_cell) {
        window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
        gtk_window_set_skip_taskbar_hint(GTK_WINDOW(window), TRUE);
        gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
        label = gtk_label_new(NULL);
        markup = g_markup_printf_escaped("<span size=\"large\" weight=\"bold\">%s</span>", _("The list of available download folders is now being updated.\n\nPlease wait a few seconds"));
        gtk_label_set_markup(GTK_LABEL(label), markup);
        g_free(markup);
        gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
        gtk_container_add(GTK_CONTAINER(window), label);
        gtk_widget_show(label);
        gtk_widget_show(window);
    }

    g_idle_add(config_changed_phase2, window);
}

void
config_folders_initialize(void)
{
    GtkWidget *treeConfigDownloadDirectory;
    GtkWidget *treeConfigTempDirectory;
    GtkTreeStore *storeDownload;
    GtkTreeStore *storeTemp;
    GtkCellRenderer *cellRenderer;
    GtkTreeSelection *selection;

    treeConfigDownloadDirectory = nntpgrab_gui_base_get_widget("treeConfigDownloadDirectory");
    treeConfigTempDirectory = nntpgrab_gui_base_get_widget("treeConfigTempDirectory");

    // Download directory
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeConfigDownloadDirectory), FALSE);
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeConfigDownloadDirectory));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);

    storeDownload = gtk_tree_store_new(1, G_TYPE_STRING);
    gtk_tree_view_set_model(GTK_TREE_VIEW(treeConfigDownloadDirectory), GTK_TREE_MODEL(storeDownload));

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeConfigDownloadDirectory), -1, _("Directory"), cellRenderer, "text", 0, NULL);

    g_signal_connect(G_OBJECT(treeConfigDownloadDirectory), "row-expanded", G_CALLBACK(on_row_expanded), NULL);
    g_signal_connect(G_OBJECT(selection), "changed", G_CALLBACK(on_selection_changed), GINT_TO_POINTER(TREE_DOWNLOAD_DIRECTORY));

    // Temp directory
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(treeConfigTempDirectory), FALSE);
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeConfigTempDirectory));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);

    storeTemp = gtk_tree_store_new(1, G_TYPE_STRING);
    gtk_tree_view_set_model(GTK_TREE_VIEW(treeConfigTempDirectory), GTK_TREE_MODEL(storeTemp));

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeConfigTempDirectory), -1, _("Directory"), cellRenderer, "text", 0, NULL);

    g_signal_connect(G_OBJECT(treeConfigTempDirectory), "row-expanded", G_CALLBACK(on_row_expanded), NULL);
    g_signal_connect(G_OBJECT(selection), "changed", G_CALLBACK(on_selection_changed), GINT_TO_POINTER(TREE_TEMP_DIRECTORY));

    ng_signal_connect(glue, "config_changed", NG_CALLBACK(config_changed), NULL);
}

void
config_folders_update_folders(void)
{
    if (initialized) {
        return;
    }
    initialized = TRUE;

    scroll_to_cell = TRUE;
    config_changed(glue, NULL);
}
