/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib-object.h>
#include <gmodule.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <wspiapi.h>
#include <io.h>
#undef  gai_strerror
#define gai_strerror gai_strerrorA
#define MSG_DONTWAIT 0
#else
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#ifdef HAVE_LIBPROXY
#include <proxy.h>
#endif

#include "nntpconnection.h"
#include "nntpgrab_internal.h"
#include "nntpgrab_plugin.h"
#include "download_queue.h"
#include "download_thread.h"

#ifndef WIN32
#define O_BINARY 0
#endif

// Keep track of the number of bytes received for the last 10 seconds
static struct timeval tv_last_traffic_monitor_flush = { 0, 0 };
static int bytes_received[10];
static GStaticMutex traffic_lock = G_STATIC_MUTEX_INIT;
static gboolean abort_traffic_thread = FALSE;

/* declaration for throttle.c */
void throttle_pause(struct timeval start_time, off_t xferlen, int max_bandwidth);

static void
strip_newline(char *line)
{
    if (line[strlen(line) - 1]  == '\n') {
        line[strlen(line) - 1] = '\0';
    }

    if (line[strlen(line) - 1]  == '\r') {
        line[strlen(line) - 1] = '\0';
    }
}

static int
get_status_code(const char *line)
{
    /* The status code is mentioned at the first 3 characters of the line.
     * atoi() doesn't detect errors and quits parsing after the first invalid
     * character which is enough for us */
    if (!line) {
        return -1;
    }

    return atoi(line);
}

#ifdef WIN32
static char *
WSAGetStrError(int err)
{
    LPVOID lpMsgBuf;
    char *ret;

    FormatMessageA(
        FORMAT_MESSAGE_ALLOCATE_BUFFER |
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        (DWORD) err,
        0,
        (LPTSTR) &lpMsgBuf,
        0, NULL );

    ret = g_strdup(lpMsgBuf);
    strip_newline(ret);

    LocalFree(lpMsgBuf);

    return ret;
}

int gettimeofday (struct timeval *tv, void* tz)
{
  union {
    long long ns100; /*time since 1 Jan 1601 in 100ns units */
    FILETIME ft;
  } now;

  GetSystemTimeAsFileTime (&now.ft);
  tv->tv_usec = (long) ((now.ns100 / 10LL) % 1000000LL);
  tv->tv_sec = (long) ((now.ns100 - 116444736000000000LL) / 10000000LL);
  return (0);
}

void timersub(struct timeval *a, struct timeval *b, struct timeval *res)
{
    res->tv_sec = a->tv_sec - b->tv_sec;
    res->tv_usec = a->tv_usec - b->tv_usec;
    if (res->tv_usec < 0) {
        res->tv_sec--;
        res->tv_usec += 1000000;
    }
}
#endif

gpointer
traffic_thread_func(gpointer data)
{
    int empty_buf[10];
    gboolean null_buf_sent = FALSE;
    time_t timestamp = 0;
#if 0
    int i;
#endif

    memset(empty_buf, 0, sizeof(empty_buf));

    while (!abort_traffic_thread) {
        time_t prev_timestamp;

        // Traffic monitoring
        prev_timestamp = timestamp;
        timestamp = time(NULL);

        if (timestamp > prev_timestamp) {
            time_t diff;
            int bytes_received_copy[10];

            // Calculate how many steps we need to shift
            diff = timestamp - prev_timestamp;
            if (diff > 10) {
                diff = 10;
            }

            g_static_mutex_lock(&traffic_lock);
            memcpy(bytes_received_copy, bytes_received, sizeof(bytes_received));

            // Shift all the previous values by 'diff'
            memmove(bytes_received, &bytes_received[diff], (10 - diff) * sizeof(int));
            memset(&bytes_received[10 - diff], 0, diff * sizeof(int));

            gettimeofday(&tv_last_traffic_monitor_flush, NULL);

            g_static_mutex_unlock(&traffic_lock);

            // Notify upstream about our values
            // If there hasn't been any sane data, send a empty buffer 1 time
            if (!memcmp(bytes_received_copy, empty_buf, sizeof(empty_buf))) {
                if (!null_buf_sent) {
                    null_buf_sent = TRUE;
                    nntpgrab_core_emit_traffic_monitor_update(FALSE, bytes_received_copy, timestamp);
                }
            } else {
                nntpgrab_core_emit_traffic_monitor_update(FALSE, bytes_received_copy, timestamp);
                null_buf_sent = FALSE;
            }

#if 0
            g_print("stamp = %li\n", timestamp);
            for (i = 0; i < 10; i++) {
                g_print("bytes_received[%i] = %i\n", i, bytes_received[i]);
            }
            g_print("\n");
#endif
        }

        g_usleep(G_USEC_PER_SEC * 1);
    }

    return NULL;
}

static void
update_traffic_monitor(int bytes_read)
{
    /* Update the bytes_received array */
    g_static_mutex_lock(&traffic_lock);
    bytes_received[9] += bytes_read;
    g_static_mutex_unlock(&traffic_lock);
}

static gboolean
get_proxy_settings(const char *hostname, char **proxy_host, int *proxy_port)
{
#ifdef HAVE_LIBPROXY
    static pxProxyFactory *proxy = NULL;
    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
    char hostname_with_https[256];
    char **proxies;
    char **values;
    int offset;

    g_return_val_if_fail(hostname != NULL, FALSE);
    g_return_val_if_fail(proxy_host != NULL, FALSE);
    g_return_val_if_fail(proxy_port != NULL, FALSE);

    /* initialize libproxy if it isn't already */
    g_static_mutex_lock(&mutex);
    if (proxy == NULL) {
        proxy = px_proxy_factory_new();
    }
    g_static_mutex_unlock(&mutex);

    /* libproxy requires the hostname to be prefixed with 'https://' */
    memset(&hostname_with_https, 0, sizeof(hostname_with_https));
    snprintf(hostname_with_https, sizeof(hostname_with_https) - 1, "https://%s", hostname);

    if (!proxy) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Unable to retrieve proxy information, pxProxyFactory() returned NULL");
        return FALSE;
    }

    proxies = px_proxy_factory_get_proxies(proxy, hostname_with_https);
    if (!proxies) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Unable to retrieve proxy information, px_proxy_factory_get_proxies() returned NULL");
        return FALSE;
    }

    if (!proxies[0]) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Unable to retrieve proxy information, px_proxy_factory_get_proxies() returned an empty list");
        return FALSE;
    }

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, "proxies[0] = %s\n", proxies[0]);

    if (!strcmp(proxies[0], "direct://")) {
        /* No proxy required */
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, "No proxy required for '%s'\n", hostname);
        return FALSE;
    }

    if (!strncmp(proxies[0], "http://", 7)) {
        offset = 7;
    } else if (!strncmp(proxies[0], "https://", 8)) {
        offset = 8;
    } else {
        offset = 0;
    }

    values = g_strsplit(proxies[0] + offset, ":", -1);
    if (!(values || values[0] || values[1])) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_WARNING, "Unable to retrieve proxy information, proxy '%s' could not be parsed", proxies[0]);
        return FALSE;
    }

    *proxy_host = g_strdup(values[0]);
    *proxy_port = atoi(values[1]);

    g_strfreev(values);

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, "Returned proxy '%s:%i' for host '%s'\n", *proxy_host, *proxy_port, hostname);

    return TRUE;
#else
    /* No libproxy support, assume we always have a direct connection */
    return FALSE;
#endif
}

static gboolean
perform_recv(NNTPConnectionInfo *conn, gboolean *try_again)
{
    int retval;
#ifdef WIN32
    char *errmsg = NULL;
#endif

    g_return_val_if_fail(conn != NULL, FALSE);
    g_return_val_if_fail(try_again != NULL, FALSE);

    *try_again = FALSE;

    if (conn->ssl) {
        retval = SSL_read(conn->ssl, conn->recv_buffer + conn->recv_buffer_length, sizeof(conn->recv_buffer) - conn->recv_buffer_length - 1);
        if (retval < 0) {
            switch (SSL_get_error(conn->ssl, retval)) {
                case SSL_ERROR_WANT_READ:
                case SSL_ERROR_WANT_WRITE:
                    retval = -1;
#ifdef WIN32
                    WSASetLastError(WSAEWOULDBLOCK);
#else
                    errno = EAGAIN;
#endif
                    break;

                case SSL_ERROR_SSL:
                    retval = -1;
#ifdef WIN32
                    errno = ENOSYS;
#else
                    errno = EPROTO;
#endif
                    break;

                default:
                    break;
            }
        }
    } else {
#ifdef WIN32
        /* Win32 doesn't support the recv() argument MSG_DONTWAIT...
         * work around this by performing a select() first */
        struct timeval tv;
        fd_set fds;

        FD_ZERO(&fds);
        FD_SET(conn->poll_fd.fd, &fds);

        tv.tv_sec = 0;
        tv.tv_usec = 0;

        if (select(conn->poll_fd.fd + 1, &fds, NULL, NULL, &tv) == 0) {
            *try_again = TRUE;
            return FALSE;
        }
#endif

        retval = recv(conn->poll_fd.fd, conn->recv_buffer + conn->recv_buffer_length, sizeof(conn->recv_buffer) - conn->recv_buffer_length - 1, MSG_DONTWAIT);
    }

    switch (retval) {
        case -1:            /* Some error occured */
#ifdef WIN32
            if (WSAGetLastError() == WSAEWOULDBLOCK) {
#else
            if (errno == EAGAIN) {
#endif
                /* Try again later */
                *try_again = TRUE;
                return FALSE;
            }

#ifdef WIN32
            errmsg = WSAGetStrError(WSAGetLastError());
            nntpconnection_disconnect_from_server(conn, DISCONNECT_READ_ERROR, errmsg);
            g_free(errmsg);
#else
            nntpconnection_disconnect_from_server(conn, DISCONNECT_READ_ERROR, strerror(errno));
#endif

            return FALSE;

        case 0:             /* Orderly shutdown has occured */
            conn->active_method = NULL;
            return FALSE;

        default:
            conn->recv_buffer_length += retval;

            /* Perform traffic shaping if necessary */
            if (conn->max_bandwidth > 0) {
                /* Calculate the sum of the number of bytes received in the last 10 seconds */
                int num_bytes_received = 0;

                g_static_mutex_lock(&traffic_lock);
                num_bytes_received = bytes_received[9];
                g_static_mutex_unlock(&traffic_lock);

                num_bytes_received += retval;
                throttle_pause(tv_last_traffic_monitor_flush, num_bytes_received, conn->max_bandwidth);
            }

            update_traffic_monitor(retval);

            break;
    }

//    nntpgrab_core_emit_debug_message(FALSE, "Received data from socket %i: %s", conn->poll_fd.fd, data);

    return TRUE;
}

static gboolean
nntpconnection_read_msg(NNTPConnectionInfo *conn, gboolean read_line, int max_length, void *data, int *data_length, gboolean *more_data_ready)
{
    int len;
    gboolean try_again = FALSE;

    g_return_val_if_fail(conn != NULL, FALSE);
    g_return_val_if_fail(max_length > 0, FALSE);
    g_return_val_if_fail(data != NULL, FALSE);
    /* NOTE: data_length MIGHT be NULL */
    /* NOTE: more_data_ready MIGHT be NULL */

    if (more_data_ready) {
        *more_data_ready = FALSE;
    }

    /* Avoid a buffer overflow */
    /* This situation can happen when a line is really long */
    if (conn->recv_buffer_length >= sizeof(conn->recv_buffer) - 1) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "nntpconnection_read_msg: Very long line received from server. Ignoring");
        memset(conn->recv_buffer, 0, sizeof(conn->recv_buffer));
        conn->recv_buffer_length = 0;
    }

    if (!perform_recv(conn, &try_again) && (!try_again || conn->recv_buffer_length == 0)) {
        return FALSE;
    }

    if (read_line) {
        char *newline = strstr(conn->recv_buffer, "\n");

        if (!newline) {
            ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "nntpconnection_read_msg: Expected newline but didn't find any. contents = %s", conn->recv_buffer);
            return FALSE;
        }

        len = newline - conn->recv_buffer + 1;

        if (len > max_length) {
            /* We're going to have to trim some data */
            ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "nntpconnection_read_msg: Buffer isn't large enough, some data will be trimmed. contents = %s", conn->recv_buffer);
            memcpy(data, conn->recv_buffer, max_length);

            if (data_length) {
                *data_length = max_length;
            }
        } else {
            memcpy(data, conn->recv_buffer, len);

            if (data_length) {
                *data_length = len;
            }
        }

        if (len == conn->recv_buffer_length) {
            /* Everything in the buffer was sent */
            conn->recv_buffer_length = 0;
            memset(conn->recv_buffer, 0, sizeof(conn->recv_buffer));
        } else {
            /* Only part of the buffer was sent */
            memmove(conn->recv_buffer, conn->recv_buffer + len, sizeof(conn->recv_buffer) - len);
            conn->recv_buffer_length -= len;

            g_return_val_if_fail(conn->recv_buffer_length >= 0, FALSE);

            if (more_data_ready) {
                *more_data_ready = TRUE;
            }
        }

        return TRUE;
    }

    if (conn->recv_buffer_length > max_length) {
        len = max_length;
    } else {
        len = conn->recv_buffer_length;
    }

    memcpy(data, conn->recv_buffer, len);

    if (len == conn->recv_buffer_length) {
        /* Everything in the buffer was sent */
        conn->recv_buffer_length = 0;
        memset(conn->recv_buffer, 0, sizeof(conn->recv_buffer));
    } else {
        /* Only part of the buffer was sent */
        memmove(conn->recv_buffer, conn->recv_buffer + len, sizeof(conn->recv_buffer) - len);
        conn->recv_buffer_length -= len;

        g_return_val_if_fail(conn->recv_buffer_length >= 0, FALSE);

        if (more_data_ready) {
            *more_data_ready = TRUE;
        }
    }

    if (data_length) {
        *data_length = len;
    }

    return TRUE;
}

static gboolean
nntpconnection_send_msg(NNTPConnectionInfo *conn, void *data, int length)
{
    int retval;
    int bytes_sent = 0;

    g_return_val_if_fail(conn != NULL, FALSE);
    g_return_val_if_fail(data != NULL, FALSE);

    if (length <= 0) {
        length = strlen(data);
    }

    do {
        if (conn->ssl) {
            retval = SSL_write(conn->ssl, data + bytes_sent, length - bytes_sent);
            if (retval < 0) {
                switch (SSL_get_error(conn->ssl, retval)) {
                    case SSL_ERROR_WANT_READ:
                    case SSL_ERROR_WANT_WRITE:
                        retval = -1;
#ifdef WIN32
                        WSASetLastError(WSAEWOULDBLOCK);
#else
                        errno = EAGAIN;
#endif
                        break;

                    case SSL_ERROR_SSL:
                        retval = -1;
#ifdef WIN32
                        errno = ENOSYS;
#else
                        errno = EPROTO;
#endif
                        break;

                    default:
                        break;
                }
            }
        } else {
            retval = send(conn->poll_fd.fd, data + bytes_sent, length - bytes_sent, 0);
        }

        switch (retval) {
            case -1:            /* Some error occured */
                nntpconnection_disconnect_from_server(conn, DISCONNECT_WRITE_ERROR, strerror(errno));
                return FALSE;

            case 0:             /* Orderly shutdown has occured */
                conn->active_method = NULL;
                return FALSE;

            default:
                bytes_sent += retval;

                if (bytes_sent == length) {
                    return TRUE;
                }

                break;
        }
    } while (TRUE);

    g_return_val_if_reached(TRUE);
}

/*************************************/
/* Initialization and authentication */
/*************************************/

static void nntpconnection_login_username_sent(NNTPConnectionInfo *conn);
static void nntpconnection_login_password_sent(NNTPConnectionInfo *conn);
static void nntpconnection_login_set_mode_reader(NNTPConnectionInfo *conn);
static void nntpconnection_login_process_mode_reader(NNTPConnectionInfo *conn);
static void nntpconnection_process_body_command(NNTPConnectionInfo *conn);
static void nnpconnection_process_body_data(NNTPConnectionInfo *conn);

static void
nntpconnection_process_welcome_msg(NNTPConnectionInfo *conn)
{
    char welcome_msg[4096];

    memset(welcome_msg, 0, sizeof(welcome_msg));

    if (!nntpconnection_read_msg(conn, TRUE, sizeof(welcome_msg) - 1, welcome_msg, NULL, NULL)) {
        return;
    }

    strip_newline(welcome_msg);

    switch (get_status_code(welcome_msg)) {
        case 200:
        case 201:
            /* Welcome message is OK */
            nntpgrab_core_emit_connected(FALSE, conn->server_info.servername, conn->poll_fd.fd, welcome_msg);
            break;

        default:
            /* Welcome message is NOT ok, probably too many connections from this IP address */
            nntpconnection_disconnect_from_server(conn, DISCONNECT_TOO_MANY_CONNECTIONS, welcome_msg);

            return;
    }

    /* Do we need to send a username and password to the server? */
    if (strlen(conn->server_info.username) > 0) {
        char buf[1024];
        memset(buf, 0, sizeof(buf));

        snprintf(buf, sizeof(buf) - 1, "AUTHINFO USER %s\r\n", conn->server_info.username);
        if (!nntpconnection_send_msg(conn, buf, strlen(buf))) {
            return;
        }

        conn->active_method = nntpconnection_login_username_sent;
    } else {
        /* We don't need to log in. Prepare the connection for further communication */
        nntpconnection_login_set_mode_reader(conn);
    }
}

static void
nntpconnection_login_username_sent(NNTPConnectionInfo *conn)
{
    char buf[1024];
    memset(buf, 0, sizeof(buf));

    if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, NULL, NULL)) {
        return;
    }

    if (get_status_code(buf) != 381) {
        /* Unexpected response code. Abort the connection */
        strip_newline(buf);
        nntpconnection_disconnect_from_server(conn, DISCONNECT_LOGIN_FAILURE, buf);
        return;
    }

    if (strlen(conn->server_info.password) > 0) {
        memset(buf, 0, sizeof(buf));
        snprintf(buf, sizeof(buf) - 1, "AUTHINFO PASS %s\r\n", conn->server_info.password);
        if (!nntpconnection_send_msg(conn, buf, strlen(buf))) {
            return;
        }

        conn->active_method = nntpconnection_login_password_sent;
    } else {
        /* No password configured for this server. Try to prepare the connection for further communication */
        nntpconnection_login_set_mode_reader(conn);
    }
}

static void
nntpconnection_login_password_sent(NNTPConnectionInfo *conn)
{
    char buf[1024];
    memset(buf, 0, sizeof(buf));

    if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, NULL, NULL)) {
        return;
    }

    if (get_status_code(buf) != 281) {
        /* Unexpected response code. Abort the connection */
        strip_newline(buf);
        nntpconnection_disconnect_from_server(conn, DISCONNECT_LOGIN_FAILURE, buf);
        return;
    }

    /* Login OK, prepare the connection for further communication */
    nntpconnection_login_set_mode_reader(conn);
}

static void
nntpconnection_login_set_mode_reader(NNTPConnectionInfo *conn)
{
    if (!nntpconnection_send_msg(conn, "MODE READER\r\n", -1)) {
        return;
    }

    conn->active_method = nntpconnection_login_process_mode_reader;
}

static void
nntpconnection_login_process_mode_reader(NNTPConnectionInfo *conn)
{
    char buf[1024];
    memset(buf, 0, sizeof(buf));

    if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, NULL, NULL)) {
        return;
    }

    switch (get_status_code(buf)) {
        case 200:
        case 201:
            /* Everything's ready! */
            break;

        case 480:
            /* Some usenet servers use this to notify there are too many connections active */
            strip_newline(buf);
            nntpconnection_disconnect_from_server(conn, DISCONNECT_TOO_MANY_CONNECTIONS, buf);
            return;

        default:
            /* Unexpected response code. Abort the connection */
            strip_newline(buf);
            nntpconnection_disconnect_from_server(conn, DISCONNECT_INVALID_MSG, buf);
            return;
    }

    /* The connection is now fully operational. Check if we need to download something */
    if (conn->collection && conn->file && conn->part) {
        nntpconnection_send_body_command(conn);
    } else {
        conn->active_method = NULL;
        conn->is_idle = TRUE;
        conn->idle_start_stamp = time(NULL);
    }
}

/**********************************************************/
/* body / article                                         */
/**********************************************************/
void
nntpconnection_send_body_command(NNTPConnectionInfo *conn)
{
    char cmd[1024];

    g_return_if_fail(conn != NULL);
    g_return_if_fail(conn->collection != NULL);
    g_return_if_fail(conn->file != NULL);
    g_return_if_fail(conn->part != NULL);
    g_return_if_fail(conn->article_fd == -1);
    g_return_if_fail(strlen(conn->article_filename) > 0);

    nntpgrab_core_emit_part_download_start(FALSE, conn->server_info.servername, conn->poll_fd.fd, conn->collection->collection_name, conn->file->subject, conn->part->part_num);

    memset(cmd, 0, sizeof(cmd));
    snprintf(cmd, sizeof(cmd) - 1, "BODY %s\r\n", conn->part->message_id);
    if (!nntpconnection_send_msg(conn, cmd, strlen(cmd))) {
        return;
    }

    conn->active_method = nntpconnection_process_body_command;
    conn->article_bytes_downloaded = 0;
    conn->article_write_buffer_length = 0;
}

static void
nntpconnection_process_body_command(NNTPConnectionInfo *conn)
{
    gboolean more_data_ready = FALSE;
    char buf[1024];
    memset(buf, 0, sizeof(buf));

    g_return_if_fail(conn != NULL);
    g_return_if_fail(conn->collection != NULL);
    g_return_if_fail(conn->file != NULL);
    g_return_if_fail(conn->part != NULL);
    g_return_if_fail(conn->article_fd == -1);
    g_return_if_fail(strlen(conn->article_filename) > 0);

    if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, NULL, &more_data_ready)) {
        return;
    }

    switch (get_status_code(buf)) {
        case 222:                       /* article retrieved - body follows */
            /* Open the file where we need to save our data to */
            conn->article_fd = open(conn->article_filename, O_CREAT | O_WRONLY | O_TRUNC | O_BINARY, S_IRUSR | S_IWUSR);
            if (conn->article_fd == -1) {
                download_thread_abort_without_waiting(_("Unable to create a file named '%s': %s"), conn->article_filename, strerror(errno));
                conn->active_method = NULL;
                return;
            }

            conn->active_method = nnpconnection_process_body_data;

            if (more_data_ready) {
                conn->active_method(conn);
            }

            return;

        case 430:                       /* no such article found */
            download_queue_update_part_status(conn, conn->collection, conn->file, conn->part, conn->server_id, FALSE, FALSE, TRUE);

            //g_object_unref(conn->collection);
            //g_object_unref(conn->file);
            //g_object_unref(conn->part);
            conn->collection = NULL;
            conn->file = NULL;
            conn->part = NULL;

            conn->active_method = NULL;
            conn->is_idle = TRUE;
            conn->idle_start_stamp = time(NULL);

            return;

        case 220:                       /* article retrieved - head and body follow */
        case 221:                       /* article retrieved - head follows */
        case 223:                       /* article retrieved - request text separately */
        case 412:                       /* no newsgroup has been selected */
        case 420:                       /* no current article has been selected */
        case 423:                       /* no such article number in this group */
        default:
            /* Unexpected response code. Abort the connection */
            strip_newline(buf);
            nntpconnection_disconnect_from_server(conn, DISCONNECT_INVALID_MSG, buf);
            return;
    }
}

static void
trim_double_dots(NNTPConnectionInfo *conn, char *buf, int *length)
{
    int i;

    g_return_if_fail(conn != NULL);
    g_return_if_fail(buf != NULL);
    g_return_if_fail(length != NULL);
    g_return_if_fail(*length > 0);

    /* If the buffer ends with a newline and a dot, push it back to the recv buffer */
    /* This is necessary to avoid confusion while un-escaping double-dots */
    if (*length >= 2 && buf[*length - 2] == '\n' && buf[*length - 1] == '.') {
        g_return_if_fail(conn->recv_buffer_length == 0);

        conn->recv_buffer[0] = '\n';
        conn->recv_buffer[1] = '.';
        conn->recv_buffer_length = 2;

        buf[*length - 2] = '\0';
        buf[*length - 1] = '\0';
        *length -= 2;
    } else if (*length >= 1 && buf[*length - 1] == '\n') {
        g_return_if_fail(conn->recv_buffer_length == 0);

        conn->recv_buffer[0] = '\n';
        conn->recv_buffer_length = 1;

        buf[*length - 1] = '\0';
        *length -= 1;
    }

    /* Un-escape any double-dots */
    for (i = 2; i < *length; i++) {
        if (buf[i - 2] == '\n' && buf[i - 1] == '.' && buf[i] == '.') {
            /* double-dot detected! trim one of the dots */
            memmove(buf + i, buf + i + 1, *length - i - 1);
            (*length)--;
        }
    }
}

static void
nnpconnection_process_body_data(NNTPConnectionInfo *conn)
{
    int length = 0;
    char buf[65536];
    gboolean more_data_ready = FALSE;
    struct timeval tv;
    struct timeval tv_diff;

    g_return_if_fail(conn != NULL);
    g_return_if_fail(conn->part != NULL);
    g_return_if_fail(conn->article_fd >= 0);

    memset(buf, 0, sizeof(buf));

    /* Keep on reading data until we've found the '\r\n.\r\n' or '\n.\n' sequence */

    if (!nntpconnection_read_msg(conn, FALSE, sizeof(buf) - 1, buf, &length, &more_data_ready)) {
        return;
    }

    g_return_if_fail(length > 0);

    conn->article_bytes_downloaded += length;
    if (conn->article_bytes_downloaded > conn->part->size) {
        conn->article_bytes_downloaded = conn->part->size;
    }

    gettimeofday(&tv, NULL);
    timersub(&tv, &conn->last_article_progress_announce, &tv_diff);
    if (tv_diff.tv_sec > 0 || tv_diff.tv_usec > G_USEC_PER_SEC / 10) {
        conn->last_article_progress_announce = tv;
        nntpgrab_core_emit_part_progress_update(FALSE, conn->server_info.servername, conn->poll_fd.fd, conn->file->subject, conn->part->part_num, conn->article_bytes_downloaded, conn->part->size);
    }

    /* Do we have enough useful data? */
    if (length < 5) {
        /* Nope, push the data back in the recv buffer and try again later */

        /* If the read_msg call returned less then 5 bytes the buffer must be empty */
        g_return_if_fail(conn->recv_buffer_length == 0);

        memcpy(conn->recv_buffer, buf, length);
        conn->recv_buffer_length = length;

        return;
    }

    /* Check for the end sequence */
    if (!strncmp(buf + length - 5, "\r\n.\r\n", 5) ||
        !strncmp(buf + length - 3, "\n.\n", 3)) {

        /* End sequence found! */

        /* Calculate the number of bytes we don't need to write to disk */
        if (!strncmp(buf + length - 5, "\r\n.\r\n", 5)) {
            length -= 5;
        } else {
            length -= 3;
        }

        /* Un-escape any double-dots */
        if (length > 0) {
            trim_double_dots(conn, buf, &length);
        }

        /* Check for a lost newline which was pushed back in the recv buffer by the trim_double_dots() function */
        if (conn->recv_buffer_length == 1 && conn->recv_buffer[0] == '\n') {
            conn->recv_buffer[0] = '\0';
            conn->recv_buffer_length = 0;
        }

        /* Flush the buffers and close the file descriptor */
        if ((conn->article_write_buffer_length > 0 &&
             write(conn->article_fd, conn->article_write_buffer, conn->article_write_buffer_length) != conn->article_write_buffer_length) ||
            (length > 0 &&
             write(conn->article_fd, buf, length) != length)) {

            /* Write error! Kill the download thread */
            download_thread_abort_without_waiting(_("%s:%i Unable to write article data to file: %s"), __FILE__, __LINE__, strerror(errno));

            /* Mark the part as failed so it will be retried later */
            download_queue_update_part_status(conn, conn->collection, conn->file, conn->part, conn->server_id, FALSE, FALSE, FALSE);
        } else {
            /* Download and save succeeded */
            download_queue_update_part_status(conn, conn->collection, conn->file, conn->part, conn->server_id, TRUE, FALSE, TRUE);
        }

        close(conn->article_fd);
        conn->article_fd = -1;

        conn->article_bytes_downloaded = 0;
        conn->article_write_buffer_length = 0;

        memset(conn->article_write_buffer, 0, sizeof(conn->article_write_buffer));

        // g_object_unref(conn->collection);
        // g_object_unref(conn->file);
        // g_object_unref(conn->part);
        conn->collection = NULL;
        conn->file = NULL;
        conn->part = NULL;

        conn->active_method = NULL;
        conn->is_idle = TRUE;

        conn->idle_start_stamp = time(NULL);

        return;
    }

    /* Un-escape any double-dots */
    trim_double_dots(conn, buf, &length);

    /* Keep the data in a temporary buffer to minimize the amount of I/O */
    if (conn->article_write_buffer_length + length > sizeof(conn->article_write_buffer)) {
        /* Flush the buffer */
        if (write(conn->article_fd, conn->article_write_buffer, conn->article_write_buffer_length) != conn->article_write_buffer_length) {
            /* Write error! Kill the download thread */
            char msg[1024];

            snprintf(msg, sizeof(msg) - 1, _("%s:%i Unable to write article data to file: %s"), __FILE__, __LINE__, strerror(errno));
            download_thread_abort_without_waiting("%s", msg);

            /* Mark the part as failed so it will be retried later */
            download_queue_update_part_status(conn, conn->collection, conn->file, conn->part, conn->server_id, FALSE, FALSE, FALSE);

            nntpconnection_disconnect_from_server(conn, DISCONNECT_WRITE_ERROR, msg);

            return;
        }

        memset(conn->article_write_buffer, 0, sizeof(conn->article_write_buffer));
        conn->article_write_buffer_length = 0;
    }

    memcpy(conn->article_write_buffer + conn->article_write_buffer_length, buf, length);
    conn->article_write_buffer_length += length;

    /* Is there more data to read? */
    if (more_data_ready) {
        nnpconnection_process_body_data(conn);
    }
}

/*****************************/
/* generic socket operations */
/*****************************/
static SSL *
prepare_ssl_connection(int conn_id, char **errmsg)
{
    SSL_CTX *ctx;
    SSL *ssl;
    SSL_METHOD *meth;
    int err;
#if 0
    char *str;
    X509 *server_cert;
#endif
    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;

    g_static_mutex_lock(&mutex);

    SSLeay_add_ssl_algorithms();
    meth = SSLv23_client_method();
    SSL_load_error_strings();
    g_static_mutex_unlock(&mutex);

    ctx = SSL_CTX_new (meth);
    if (!ctx) {
        if (errmsg) {
            *errmsg = g_strdup("SSL_CTX_new FAILED");
        }

        return NULL;
    }

    /* Don't verify the certificate as several usenet providers (like Eweka) use self-signed certificates */
    SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, NULL);

    ssl = SSL_new (ctx);
    SSL_CTX_free(ctx);

    if (!ssl) {
        if (errmsg) {
            *errmsg = g_strdup("SSL_new FAILED");
        }

        return NULL;
    }

    SSL_set_fd (ssl, conn_id);
    err = SSL_connect (ssl);
    if (err <= 0) {
        if (errmsg) {
            *errmsg = g_strdup_printf("%s", ERR_error_string(ERR_get_error(), NULL));
        }

        return NULL;
    }

#if 0
    printf ("SSL connection using %s\n", SSL_get_cipher (ssl));
    server_cert = SSL_get_peer_certificate (ssl);
    if (server_cert) {
        printf ("Server certificate:\n");

        str = X509_NAME_oneline (X509_get_subject_name (server_cert),0,0);
        if (str) {
            printf ("\t subject: %s\n", str);
            OPENSSL_free (str);
        }

        str = X509_NAME_oneline (X509_get_issuer_name (server_cert),0,0);
        if (str) {
            printf ("\t issuer: %s\n", str);
            OPENSSL_free (str);
        }

        /* We could do all sorts of certificate verification stuff here before deallocating the certificate. */

        X509_free (server_cert);
    }
#endif

    return ssl;
}

/********************/
/* proxy operations */
/********************/
static void
nntpconnection_process_proxy_connect_command(NNTPConnectionInfo *conn)
{
    char buf[128];
    int length = 0;
    char *ptr;

    g_return_if_fail(conn != NULL);

    memset(buf, 0, sizeof(buf));
    if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, &length, NULL)) {
        return;
    }

    /* buf should now contain something like 'HTTP/1.0 200 Connection established' */
    ptr = strstr(buf, " ");
    if (get_status_code(ptr + 1) != 200) {
        /* Unknown response code, disconnect */
        strip_newline(buf);
        nntpconnection_disconnect_from_server(conn, DISCONNECT_INVALID_MSG, buf);
        return;
    }

    /* Keep on reading until we've found a blank line */
    do {
        memset(&buf, 0, sizeof(buf));
        if (!nntpconnection_read_msg(conn, TRUE, sizeof(buf) - 1, buf, &length, NULL)) {
            return;
        }

        strip_newline(buf);
    } while (strlen(buf) > 0);

    /* From now on we can talk the regular NNTP protocol */
    conn->active_method = nntpconnection_process_welcome_msg;

    /* Prepare the SSL connection if necessary */
    if (conn->server_info.use_ssl) {
        char *errmsg = NULL;

        if ((conn->ssl = prepare_ssl_connection(conn->poll_fd.fd, &errmsg)) == NULL) {
            nntpconnection_disconnect_from_server(conn, DISCONNECT_ERROR_SSL_INITIALISE, errmsg);
            g_free(errmsg);
            return;
        }
    }
}

NNTPConnectionErrCode
nntpconnection_connect_to_server(NNTPConnectionInfo *conn, char **errmsg)
{
    static gboolean traffic_monitor_started = FALSE;
    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
    struct addrinfo hints, *res, *ressave;
    int n;
    char *str_port;
    int last_errno;
#ifdef WIN32
    int tv;
    u_long ioctlArg;
#else
    struct timeval tv;
    int mode;
#endif
    NNTPDisconnectType disconnect_type;
    char *proxy_host = NULL;
    int proxy_port = 0;

    g_return_val_if_fail(conn != NULL, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(conn->poll_fd.fd == -1, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(strlen(conn->server_info.hostname) > 0, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);
    g_return_val_if_fail(conn->server_info.port > 0, NNTP_CONNECTION_ERROR_INVALID_ARGUMENT);

    g_static_mutex_lock(&mutex);
    if (!traffic_monitor_started) {
        traffic_monitor_started = TRUE;
        memset(&bytes_received, 0, sizeof(bytes_received));

        /* TODO: implement a nice clean up */
        abort_traffic_thread = FALSE;
        /*traffic_thread = */g_thread_create(traffic_thread_func, NULL, TRUE, NULL);
    }
    g_static_mutex_unlock(&mutex);

    conn->is_idle = FALSE;

    get_proxy_settings(conn->server_info.hostname, &proxy_host, &proxy_port);

    memset(&hints, 0, sizeof(struct addrinfo));

    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    if (proxy_host) {
        str_port = g_strdup_printf("%i", proxy_port);
        n = getaddrinfo(proxy_host, str_port, &hints, &res);
        g_free(str_port);
    } else {
        str_port = g_strdup_printf("%i", conn->server_info.port);
        n = getaddrinfo(conn->server_info.hostname, str_port, &hints, &res);
        g_free(str_port);
    }

    if (n != 0) {
        if (errmsg) {
            *errmsg = g_strdup(gai_strerror(n));
        }
        return NNTP_CONNECTION_ERROR_HOST_NOT_FOUND;
    }

    ressave = res;
    last_errno = 0;

    while (res) {
        int ret;

        conn->poll_fd.fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);

#ifdef WIN32
        if ((conn->poll_fd.fd == INVALID_SOCKET)) {
#else
        if ((conn->poll_fd.fd == -1)) {
#endif
            // The socket couldn't be created
            // Save the errno and try the next item in the list
#ifdef WIN32
            last_errno = WSAGetLastError();
#else
            last_errno = errno;
#endif
            res = res->ai_next;
            conn->poll_fd.fd = -1;

            if (errmsg) {
                if (*errmsg) {
                    g_free(*errmsg);
                }

#ifdef WIN32
                *errmsg = WSAGetStrError(last_errno);
#else
                *errmsg = g_strdup(strerror(last_errno));
#endif
            }

            continue;
        }

        // Set the connection timeout on the socket
#ifdef WIN32
        tv = 5000;   /* F*cking Winsock uses an integer for the timeout and it's value is in milliseconds */

        setsockopt(conn->poll_fd.fd, SOL_SOCKET, SO_RCVTIMEO, (char*) &tv, sizeof(tv));
        setsockopt(conn->poll_fd.fd, SOL_SOCKET, SO_SNDTIMEO, (char*) &tv, sizeof(tv));
#else
        tv.tv_sec = 5;
        tv.tv_usec = 0;

        setsockopt(conn->poll_fd.fd, SOL_SOCKET, SO_RCVTIMEO, (const void*) &tv, sizeof(tv));
        setsockopt(conn->poll_fd.fd, SOL_SOCKET, SO_SNDTIMEO, (const void*) &tv, sizeof(tv));
#endif

        // Try to connect to the server
        nntpgrab_core_emit_connecting(FALSE, conn->server_info.servername, conn->poll_fd.fd);

        // Mark the socket as non-blocking
#ifdef WIN32
        ioctlArg = 1;
        ioctlsocket(conn->poll_fd.fd, FIONBIO, &ioctlArg);
#else
        mode = fcntl(conn->poll_fd.fd, F_GETFL, 0);
        fcntl(conn->poll_fd.fd, F_SETFL, mode | O_NONBLOCK);
#endif

        ret = connect(conn->poll_fd.fd, res->ai_addr, (int) res->ai_addrlen);
        if (ret == 0) {
            // Connection succesfull
            last_errno = 0;
            break;
#ifndef WIN32
        } else if (errno == EINPROGRESS) {
            // Wait for at most 5 seconds
            socklen_t len;

            time_t now = time(NULL);
            do {
                // According to http://cr.yp.to/docs/connect.html there are various
                // ways to detect if a socket is connected. The getpeername() method
                // should be the most portable one
                if (res->ai_family == AF_INET6) {
                    struct sockaddr_in6 name;
                    len = sizeof(name);
                    if (getpeername(conn->poll_fd.fd, (struct sockaddr*) &name, &len) == 0) {
                        errno = 0;
                        break;
                    }
                } else {
                    struct sockaddr_in name;
                    len = sizeof(name);
                    if (getpeername(conn->poll_fd.fd, (struct sockaddr*) &name, &len) == 0) {
                        errno = 0;
                        break;
                    }
                }

                g_usleep(G_USEC_PER_SEC / 10);

                if (download_thread_get_state() != SCHEDULAR_STATE_RUNNING) {
                    break;
                }
            } while (now + 5 > time(NULL));

            if (errno == 0) {
                last_errno = 0;
                break;
            }
#else
        } else if (WSAGetLastError() == WSAEWOULDBLOCK) {
            // And Windows again has different behaviour....
            struct timeval tv;
            fd_set send_fds;
            int len;

            FD_ZERO(&send_fds);
            FD_SET(conn->poll_fd.fd, &send_fds);

            tv.tv_sec = 5;
            tv.tv_usec = 0;

            len = select(conn->poll_fd.fd + 1, NULL, &send_fds, NULL, &tv);
            if (len <= 0) {
                // Read timeout or some other error
                if (len == 0) {
                    errno = WSAETIMEDOUT;
                }
            } else {
                // Connection is made
                last_errno = 0;
                break;
            }
#endif
        }

        // Connection could not be made, save the errno
#ifndef WIN32
        if (errno == EINPROGRESS) {
            // The connection could not be established within the connect time limit.
            // This provides a more clear error message
            last_errno = ETIMEDOUT;
        } else
#endif
        {
            last_errno = errno;
        }

        if (errmsg) {
            if (*errmsg) {
                g_free(*errmsg);
            }

#ifdef WIN32
            *errmsg = WSAGetStrError(last_errno);
#else
            *errmsg = g_strdup(strerror(last_errno));
#endif
        }

#if WIN32
        if (last_errno == WSAETIMEDOUT) {
#else
        if (last_errno == ETIMEDOUT) {
#endif
            disconnect_type = DISCONNECT_CONNECT_TIMEOUT;
        } else {
            disconnect_type = DISCONNECT_CONNECTION_REFUSED;
        }

        nntpconnection_disconnect_from_server(conn, disconnect_type, strerror(last_errno));

        res = res->ai_next;
    }

    if (conn->poll_fd.fd == -1) {
        freeaddrinfo(ressave);

        switch (last_errno) {
#if WIN32
            case WSAETIMEDOUT:
#else
            case ETIMEDOUT:
#endif
                return NNTP_CONNECTION_ERROR_CONNECTION_TIMEOUT;

            default:
                return NNTP_CONNECTION_ERROR_CONNECTION_REFUSED;
        };
    }

    // Mark the socket as blocking
#ifdef WIN32
        ioctlArg = 0;
        ioctlsocket(conn->poll_fd.fd, FIONBIO, &ioctlArg);
#else
        mode = fcntl(conn->poll_fd.fd, F_GETFL, 0);
        fcntl(conn->poll_fd.fd, F_SETFL, mode  ^ O_NONBLOCK);
#endif

    freeaddrinfo(ressave);

    if (conn->server_info.use_ssl && !proxy_host) {
        if ((conn->ssl = prepare_ssl_connection(conn->poll_fd.fd, errmsg)) == NULL) {
            nntpconnection_disconnect_from_server(conn, DISCONNECT_ERROR_SSL_INITIALISE, *errmsg);
            return NNTP_CONNECTION_ERROR_SSL_INITIALISE;
        }
    }

    if (proxy_host) {
        char cmd[128];

        memset(&cmd, 0, sizeof(cmd));
        snprintf(cmd, sizeof(cmd) - 1, "CONNECT %s:%i HTTP/1.1\r\nHost: %s:%i\r\n\r\n", conn->server_info.hostname, conn->server_info.port, conn->server_info.hostname, conn->server_info.port);
        send(conn->poll_fd.fd, cmd, strlen(cmd), 0);

        g_free(proxy_host);

        conn->active_method = nntpconnection_process_proxy_connect_command;
    } else {
        conn->active_method = nntpconnection_process_welcome_msg;
    }
    memset(conn->recv_buffer, 0, sizeof(conn->recv_buffer));
    conn->recv_buffer_length = 0;
    conn->last_activity_stamp = time(NULL);

    return NNTP_CONNECTION_ERROR_NONE;
    //return login(username, password, errmsg, conn->poll_fd.fd, data);
}

void
nntpconnection_disconnect_from_server(NNTPConnectionInfo *conn, NNTPDisconnectType disconnect_type, const char *reason)
{
    g_return_if_fail(conn != NULL);
    g_return_if_fail(conn->poll_fd.fd >= 0);

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Request received to disconnect connection %i, disconnect_type = %i, reason = %s", conn->poll_fd.fd, disconnect_type, reason);

#ifdef WIN32
    /* DisconnectEx(conn->poll_fd.fd, NULL, 0, 0);   *  MinGW doesn't have DisconnectEx exported... */
    shutdown(conn->poll_fd.fd, SD_BOTH);            /* The shutdown() function is broken on Win32, see
                                                     * http://msdn.microsoft.com/en-us/library/ms738547%28VS.85%29.aspx
                                                     * for details*/
    closesocket(conn->poll_fd.fd);
#else
    shutdown(conn->poll_fd.fd, SHUT_RDWR);
    close(conn->poll_fd.fd);
#endif

    if (conn->collection && conn->file && conn->part) {
        /* Update the file status */
        download_queue_update_part_status(conn, conn->collection, conn->file, conn->part, conn->server_id, FALSE, FALSE, FALSE);
    }

    nntpgrab_core_emit_disconnect(FALSE, conn->server_info.servername, conn->poll_fd.fd, disconnect_type, reason);

    conn->poll_fd.fd = -1;
    conn->is_idle = TRUE;
    if (disconnect_type == DISCONNECT_NORMAL) {
        conn->last_disconnect_stamp = 0;
    } else {
        conn->last_disconnect_stamp = time(NULL);
    }
    conn->last_activity_stamp = 0;

    if (conn->collection) {
        //g_object_unref(conn->collection);
        conn->collection = NULL;
    }

    if (conn->file) {
        //g_object_unref(conn->file);
        conn->file = NULL;
    }

    if (conn->part) {
        //g_object_unref(conn->part);
        conn->part = NULL;
    }

    if (conn->article_fd >= 0) {
        close(conn->article_fd);
        conn->article_fd = -1;
    }
}

void
nntpconnection_process_socket_activity(NNTPConnectionInfo *conn)
{
    g_return_if_fail(conn != NULL);

    if (!conn->active_method) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "NNTP connection to server '%s' (socket_id %i) is in an undefined state", conn->server_info.servername, conn->poll_fd.fd);
        nntpconnection_disconnect_from_server(conn, DISCONNECT_UNEXPECTED, _("NNTP Connection is in an undefined state"));
        return;
    }

    conn->active_method(conn);
    conn->last_activity_stamp = time(NULL);
}
