/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <glib-object.h>
#include "nntpgrab.h"
#include "nntpgrab_internal.h"
#include "nntpgrab_utils.h"
#include "marshalers.h"
#include "configuration.h"
#include "nntpconnection.h"
#include "collection_alloc.h"
#include "download_thread.h"
#include "download_queue.h"
#include "decoder_thread.h"
#include "plugins.h"

#ifdef WIN32
#define tempnam _tempnam
#endif

#define NNTPGRAB_TYPE_CORE              (core_get_type ())
#define NNTPGRAB_CORE(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), NNTPGRAB_TYPE_CORE, NNTPGrabCore))
#define NNTPGRAB_CORE_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), NNTPGRAB_TYPE_CORE, NNTPGrabCoreClass))
#define IS_NNTPGRAB_CORE(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), NNTPGRAB_TYPE_CORE))
#define IS_NNTPGRAB_CORE_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), NNTPGRAB_TYPE_CORE))
#define NNTPGRAB_CORE_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), NNTPGRAB_TYPE_CORE, NNTPGrabCoreClass))

struct NNTPGrabCore
{
    GObject parent;

    Configuration *config;
    gboolean is_finalizing;
};

static NNTPGrabCore *core = NULL;

struct NNTPGrabCoreClass
{
    GObjectClass parent;
};

enum
{
    CONFIG_CHANGED_SIGNAL,
    PART_DOWNLOAD_START_SIGNAL,
    PART_DONE_SIGNAL,
    PART_FAILED_SIGNAL,
    PART_PROGRESS_UPDATE_SIGNAL,
    TRAFFIC_MONITOR_UPDATE_SIGNAL,
    COLLECTION_ADDED_SIGNAL,
    COLLECTION_REMOVED_SIGNAL,
    COLLECTION_MODIFIED_SIGNAL,
    FILE_ADDED_SIGNAL,
    FILE_REMOVED_SIGNAL,
    FILE_DOWNLOAD_STATE_UPDATE_SIGNAL,
    FILE_STATE_CHANGED_SIGNAL,
    CONNECTION_CONNECTING_SIGNAL,
    CONNECTION_CONNECTED_SIGNAL,
    CONNECTION_DISCONNECT_SIGNAL,
    SCHEDULAR_STATE_CHANGED_SIGNAL,
    FATAL_ERROR_SIGNAL,
    WARNING_MESSAGE_SIGNAL,
    LOG_MESSAGE_SIGNAL,
    TASK_MOVED_SIGNAL,
    COLLECTION_MOVED_SIGNAL,
    COLLECTION_DOWNLOADED_SIGNAL,
    ALL_DOWNLOADS_COMPLETED_SIGNAL,
    PLUGIN_LOADED_SIGNAL,
    PLUGIN_UNLOADED_SIGNAL,
    PLUGIN_EVENT_SIGNAL,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(NNTPGrabCore, core, G_TYPE_OBJECT)

static void
core_init (NNTPGrabCore *obj)
{
    obj->is_finalizing = FALSE;
}

static void
core_finalize (GObject *obj)
{
    char *errmsg = NULL;
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    core->is_finalizing = TRUE;

    if (download_thread_get_state() == SCHEDULAR_STATE_RUNNING) {
        download_thread_stop(NULL);
    }

    if (!download_queue_destroy(&errmsg)) {
        g_print(__FILE__ ":%i Unable to clean up the download queue: %s\n", __LINE__, errmsg);
        g_free(errmsg);
    }

    decoder_thread_destroy();
    plugins_destroy();

    core_signals_clear_stack();

    if (core->config) {
        configuration_destroy(core->config);
    }
}

static void
core_class_init (NNTPGrabCoreClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = core_finalize;

#define create_signal(a,b,c,d,e,f,g,h,i, j...)                    \
            g_signal_new(a,b,c,d,e,f,g,h,i, ## j);                \
            ng_plugin_create_internal_event(a, g, h, i, ## j);

    signals[CONFIG_CHANGED_SIGNAL] =      create_signal("config_changed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__VOID,
                                                        G_TYPE_NONE, 0);

    signals[PART_DOWNLOAD_START_SIGNAL] = create_signal("part_download_start",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT,
                                                        G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);

    signals[PART_DONE_SIGNAL] =           create_signal("part_done",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT_INT,
                                                        G_TYPE_NONE, 6, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    signals[PART_FAILED_SIGNAL] =         create_signal("part_failed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT_INT_BOOLEAN,
                                                        G_TYPE_NONE, 7, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_BOOLEAN);

    signals[PART_PROGRESS_UPDATE_SIGNAL] = create_signal("part_progress_update",
                                                         G_OBJECT_CLASS_TYPE (klass),
                                                         G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                         0,
                                                         NULL, NULL,
                                                         nntpgrab_marshal_VOID__STRING_INT_STRING_INT_INT_INT,
                                                         G_TYPE_NONE, 6, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT);

    signals[TRAFFIC_MONITOR_UPDATE_SIGNAL] = create_signal("traffic_monitor_update",
                                                         G_OBJECT_CLASS_TYPE (klass),
                                                         G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                         0,
                                                         NULL, NULL,
                                                         nntpgrab_marshal_VOID__INT_INT_INT_INT_INT_INT_INT_INT_INT_INT_INT64_DOUBLE,
                                                         G_TYPE_NONE, 12, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT64, G_TYPE_DOUBLE);

    signals[COLLECTION_ADDED_SIGNAL] =    create_signal("collection_added",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING,
                                                        G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    signals[COLLECTION_REMOVED_SIGNAL] =  create_signal("collection_removed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING,
                                                        G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[COLLECTION_REMOVED_SIGNAL] =  create_signal("collection_modified",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING,
                                                        G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    signals[FILE_ADDED_SIGNAL] =          create_signal("file_added",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING_STRING_UINT64_UINT64_UINT64_UINT64_INT_INT_POINTER,
                                                        G_TYPE_NONE, 10, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_INT, G_TYPE_INT, G_TYPE_POINTER);

    signals[FILE_REMOVED_SIGNAL] =        create_signal("file_removed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING_UINT64_UINT64,
                                                        G_TYPE_NONE, 4, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT64, G_TYPE_UINT64);

    signals[FILE_DOWNLOAD_STATE_UPDATE_SIGNAL] =  create_signal("file_download_state_update",
                                                                G_OBJECT_CLASS_TYPE (klass),
                                                                G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                                0,
                                                                NULL, NULL,
                                                                nntpgrab_marshal_VOID__STRING_STRING_INT_INT_INT_UINT64_UINT64_UINT64_UINT64,
                                                                G_TYPE_NONE, 9, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_UINT64);

    signals[FILE_STATE_CHANGED_SIGNAL] =  create_signal("file_state_changed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING_STRING_INT_INT,
                                                        G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    signals[CONNECTION_CONNECTING_SIGNAL] =    create_signal("connection_connecting",
                                                             G_OBJECT_CLASS_TYPE (klass),
                                                             G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                             0,
                                                             NULL, NULL,
                                                             nntpgrab_marshal_VOID__STRING_INT,
                                                             G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_INT);

    signals[CONNECTION_CONNECTED_SIGNAL] =     create_signal("connection_connected",
                                                             G_OBJECT_CLASS_TYPE (klass),
                                                             G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                             0,
                                                             NULL, NULL,
                                                             nntpgrab_marshal_VOID__STRING_INT_STRING,
                                                             G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING);

    signals[CONNECTION_DISCONNECT_SIGNAL] = create_signal("connection_disconnect",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_INT_ENUM_STRING,
                                                          G_TYPE_NONE, 4, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_STRING);


    signals[SCHEDULAR_STATE_CHANGED_SIGNAL] = create_signal("schedular_state_changed",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__INT_STRING,
                                                          G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_STRING);


    signals[FATAL_ERROR_SIGNAL] =           create_signal("fatal_error",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING,
                                                          G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[WARNING_MESSAGE_SIGNAL] =       create_signal("warning_message",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING,
                                                          G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[LOG_MESSAGE_SIGNAL] =           create_signal("log_message",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_INT_STRING,
                                                          G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING);

    signals[TASK_MOVED_SIGNAL] =            create_signal("task_moved",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_STRING_INT_INT,
                                                          G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    signals[COLLECTION_MOVED_SIGNAL] =      create_signal("collection_moved",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_INT_INT,
                                                          G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);

    signals[COLLECTION_DOWNLOADED_SIGNAL] =   create_signal("collection_downloaded",
                                                            G_OBJECT_CLASS_TYPE (klass),
                                                            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                            0,
                                                            NULL, NULL,
                                                            nntpgrab_marshal_VOID__STRING,
                                                            G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[ALL_DOWNLOADS_COMPLETED_SIGNAL] = create_signal("all_downloads_completed",
                                                            G_OBJECT_CLASS_TYPE (klass),
                                                            G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                            0,
                                                            NULL, NULL,
                                                            nntpgrab_marshal_VOID__VOID,
                                                            G_TYPE_NONE, 0);

    signals[PLUGIN_LOADED_SIGNAL] =         create_signal("plugin_loaded",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_BOOL,
                                                          G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);

    signals[PLUGIN_UNLOADED_SIGNAL] =       create_signal("plugin_unloaded",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING,
                                                          G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[PLUGIN_EVENT_SIGNAL] =          create_signal("plugin_event",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_STRING_BOXED,
                                                          G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRV);
}

NNTPGrabCore *
get_core(void)
{
    if (core && core->is_finalizing) {
        return NULL;
    }

    return core;
}

static gboolean
on_config_changed(Configuration *config, gpointer data)
{
    nntpgrab_core_emit_config_changed(FALSE);
    return FALSE;
}

/***************************************************************/
/*                                                             */
/*                    Public methods                           */
/*                                                             */
/***************************************************************/

NNTPGrabCore *
nntpgrab_core_init(int version, char **err, char **warnings)
{
    static gboolean initialized = FALSE;
#ifdef WIN32
    char *installation_path;
#endif

    if (initialized) {
        if (err) {
            *err = g_strdup_printf(_("NNTPGrab Core already initialized"));
        }
        return NULL;
    }

    if (!g_thread_supported ()) g_thread_init (NULL);
    g_type_init();

    // Check if the API version matches
    if (version != NNTPGRAB_API_VERSION) {
        if (err) {
            *err = g_strdup_printf(_("NNTPGrab Core API mismatch (Core API version = %i, frontend's version = %i)"), NNTPGRAB_API_VERSION, version);
        }
        return NULL;
    }

#ifdef WIN32
    // Detect the installation directory on Win32
    installation_path = g_win32_get_package_installation_directory_of_module(NULL);
    g_return_val_if_fail(installation_path != NULL, NULL);

    if (!g_getenv("NNTPGRAB_LIBDIR")) {
        char *path = g_strdup_printf("%s\\lib\\nntpgrab", installation_path);
        g_setenv("NNTPGRAB_LIBDIR", path, TRUE);
        g_free(path);
    }

    if (!g_getenv("NNTPGRAB_WWWDIR")) {
        char *path = g_strdup_printf("%s\\webinterface", installation_path);
        g_setenv("NNTPGRAB_WWWDIR", path, TRUE);
        g_free(path);
    }

    g_free(installation_path);
#endif

    if (!plugins_initialize(err)) {
        g_object_unref(core);
        return NULL;
    }

    core = g_object_new(NNTPGRAB_TYPE_CORE, NULL);

    if (!(core->config = configuration_new(err, warnings))) {
        g_object_unref(core);
        return NULL;
    }

    g_signal_connect(core->config, "config_changed", G_CALLBACK(on_config_changed), NULL);

    if (!(download_queue_init(core->config, err))) {
        g_object_unref(core);
        return NULL;
    }

    if (!plugins_load(core->config, err)) {
        g_object_unref(core);
        return NULL;
    }

    /* Try to load the 'Unpack' and 'Auto-import' plugins if they're available */
    plugins_load_plugin_by_name("Unpack", NULL);
    plugins_load_plugin_by_name("Auto-import", NULL);

    decoder_thread_initialize(core->config);

#if 0
    if (!par2_initialize(err)) {
        g_object_unref(core);
        return NULL;
    }

    if (!unpack_initialize(err)) {
        g_object_unref(core);
        return NULL;
    }
#endif

    initialized = TRUE;

    return core;
}

void
nntpgrab_core_cleanup(NNTPGrabCore *obj)
{
    g_object_unref(obj);
}

NGList *
nntpgrab_core_config_get_avail_servers(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->config);

    return (NGList*) configuration_get_avail_servers(core->config);
}

void
nntpgrab_core_config_free_avail_servers(NNTPGrabCore *obj, NGList *servers)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->config);

    configuration_free_avail_servers(core->config, (GList*) servers);
}

ngboolean
nntpgrab_core_config_get_server_info(NNTPGrabCore *obj, const char *servername, ConfigServer *ret)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    ConfigServer *config;

    g_assert(core->config);

    config = configuration_get_server_info(core->config, servername);
    if (config) {
        memcpy(ret, config, sizeof(ConfigServer));
        g_slice_free(ConfigServer, config);
        return TRUE;
    } else {
        return FALSE;
    }
}

ngboolean
nntpgrab_core_config_add_server(NNTPGrabCore *obj, ConfigServer new_server, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    gboolean retval;
    char *errmsg_save = NULL;

    g_assert(core->config);

    retval = configuration_add_server(core->config, new_server, errmsg);

    if (!configuration_save(core->config, &errmsg_save)) {
        nntpgrab_core_emit_fatal_error(FALSE, errmsg_save);
        g_free(errmsg_save);
    }

    return retval;
}

ngboolean
nntpgrab_core_config_del_server(NNTPGrabCore *obj, const char *servername, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    gboolean retval;
    char *errmsg_save = NULL;

    g_assert(core->config);

    retval = configuration_del_server(core->config, servername, errmsg);

    if (!configuration_save(core->config, &errmsg_save)) {
        nntpgrab_core_emit_fatal_error(FALSE, errmsg_save);
        g_free(errmsg_save);
    }

    return retval;
}

ngboolean
nntpgrab_core_config_edit_server(NNTPGrabCore *obj, const char *servername, ConfigServer server, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    gboolean retval;
    char *errmsg_save = NULL;

    g_assert(core->config);

    retval = configuration_edit_server(core->config, servername, server, errmsg);

    if (!configuration_save(core->config, &errmsg_save)) {
        nntpgrab_core_emit_fatal_error(FALSE, errmsg_save);
        g_free(errmsg_save);
    }

    return retval;
}

ConfigOpts
nntpgrab_core_config_get_opts(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->config);

    return configuration_get_opts(core->config);
}

void
nntpgrab_core_config_set_opts(NNTPGrabCore *obj, ConfigOpts opts)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *errmsg = NULL;

    g_assert(core->config);

    configuration_set_opts(core->config, opts);

    if (!configuration_save(core->config, &errmsg)) {
        nntpgrab_core_emit_fatal_error(FALSE, errmsg);
        g_free(errmsg);
    }
}

ngboolean
nntpgrab_core_config_get_folder_listing(NNTPGrabCore *obj, const char *parent, NGList **folders)
{
    return nntpgrab_utils_get_folder_listing(parent, folders);
}

void
nntpgrab_core_config_free_folder_listing(NNTPGrabCore *obj, NGList *folders)
{
    nntpgrab_utils_free_folder_listing(folders);
}

gboolean
nntpgrab_core_schedular_start(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);

    return download_thread_start(core->config);
}

static gpointer
schedular_stop_func(gpointer data)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(data);

    g_return_val_if_fail(core != NULL, NULL);

    download_thread_stop(NULL);

    return NULL;
}

ngboolean
nntpgrab_core_schedular_stop(NNTPGrabCore *obj, ngboolean wait)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);

    if (wait) {
        return download_thread_stop(NULL);
    } else {
        // Is the schedular queue already stopped?
        if (download_thread_get_state() != SCHEDULAR_STATE_RUNNING) {
            return FALSE;
        }

        g_thread_create(schedular_stop_func, obj, FALSE, NULL);

        return TRUE;
    }
}

NGSchedularState
nntpgrab_core_schedular_get_state(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, SCHEDULAR_STATE_STOPPED);

    return (download_thread_get_state());
}

ngboolean
nntpgrab_core_schedular_add_task_to_queue(NNTPGrabCore *obj, const char *collection_name, const char *subject, const char *poster, time_t stamp, nguint64 file_size, NGList *groups, NGList *parts, char **errmsg)
{
    return download_queue_add_file_to_queue(collection_name, subject, poster, stamp, file_size, groups, parts, errmsg);
}

ngboolean
nntpgrab_core_schedular_del_task_from_queue(NNTPGrabCore *obj, const char *collection_name, const char *subject, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(collection_name != NULL, FALSE);
    /* NOTE: subject MIGHT be NULL ! */

    return download_queue_del_file_from_queue(collection_name, subject, errmsg);
}

ngboolean
nntpgrab_core_schedular_restart_task(NNTPGrabCore *obj, const char *collection_name, const char *subject, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(collection_name != NULL, FALSE);
    /* NOTE: subject MIGHT be NULL ! */

    return download_queue_restart_file(collection_name, subject, errmsg);
}

ngboolean
nntpgrab_core_schedular_save_queue(NNTPGrabCore *obj, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);

    return download_queue_save(errmsg);
}

void
nntpgrab_core_schedular_foreach_task(NNTPGrabCore *obj, ForeachCollectionFunc collection_func, ForeachFileFunc file_func, ForeachGroupFunc group_func, void *data)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_return_if_fail(core != NULL);

    download_queue_foreach_file(collection_func, file_func, group_func, data);
}

ngboolean
nntpgrab_core_schedular_move_task(NNTPGrabCore *obj, const char *collection_name_src, const char *subject_src, const char *collection_name_dest, int position_dest)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(collection_name_src != NULL, FALSE);
    g_return_val_if_fail(subject_src != NULL, FALSE);
    g_return_val_if_fail(collection_name_dest != NULL, FALSE);
    g_return_val_if_fail(position_dest >= -1, FALSE);

    return download_queue_move_file(collection_name_src, subject_src, collection_name_dest, position_dest);
}

ngboolean
nntpgrab_core_schedular_move_collection(NNTPGrabCore *obj, const char *collection_name, int new_position)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_return_val_if_fail(core != NULL, FALSE);
    g_return_val_if_fail(collection_name != NULL, FALSE);
    g_return_val_if_fail(new_position >= -1, FALSE);

    return download_queue_move_collection(collection_name, new_position);
}

NGList *
nntpgrab_core_plugins_get_avail_plugins(NNTPGrabCore *obj)
{
    return plugins_get_avail_plugins();
}

void
nntpgrab_core_plugins_free_avail_plugins(NNTPGrabCore *obj, NGList *plugins)
{
    plugins_free_avail_plugins(plugins);
}

ngboolean
nntpgrab_core_plugins_get_plugin_info(NNTPGrabCore *obj, const char *plugin_name, NNTPGrabPluginInfo *plugin_info)
{
    return plugins_get_plugin_info(plugin_name, plugin_info);
}

ngboolean
nntpgrab_core_plugins_load_plugin(NNTPGrabCore *obj, const char *plugin_name, char **errmsg)
{
    return plugins_load_plugin_by_name(plugin_name, errmsg);
}

ngboolean
nntpgrab_core_plugins_unload_plugin(NNTPGrabCore *obj, const char *plugin_name, char **errmsg)
{
    return plugins_unload_plugin_by_name(plugin_name, errmsg);
}

ngboolean
nntpgrab_core_plugins_set_persistent(NNTPGrabCore *obj, const char *plugin_name, ngboolean persistent)
{
    return plugins_set_persistent(plugin_name, persistent);
}

ngboolean
nntpgrab_core_embedded_server_start(NNTPGrabCore *obj, int port, char **errmsg)
{
    ngboolean retval = FALSE;
    NNTPGrabPluginInfo plugin_info;

    if (!plugins_get_plugin_info("JSON-RPC", &plugin_info)) {
        *errmsg = g_strdup_printf(_("Unable to start the embedded server as the JSON-RPC plugin could not be found"));
        return FALSE;
    }

    if (!plugin_info.is_loaded &&
        !plugins_load_plugin_by_name("JSON-RPC", errmsg)) {

        return FALSE;
    }

    ng_plugin_call(NULL, "webserver_start", port, errmsg, &retval);
    ng_plugin_call(NULL, "webserver_ignore_config_changes");

    return retval;
}
