/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _NNTPGRAB_PLUGINS_H_
#define _NNTPGRAB_PLUGINS_H_

#include "nntpgrab_types.h"
#include "nntpgrab_plugin.h"
#include "configuration.h"
#include <gmodule.h>

struct _ng_plugin_core_data {
    GModule *plugin;
    gboolean is_loaded;
    ngboolean is_required;

    char name[128];
    char version[64];
    char author[128];
    char url[256];
    char description[4096];

    GList *exported_functions;
    GList *required_functions_and_events;
    GList *exported_events;
    GList *dependencies;

    gboolean initialisation_error;
    char initialisation_errmsg[4096];

    void (*init_func) (NGPlugin *plugin_data);
    ngboolean (*load_func) (NGPlugin *plugin_data, char **errmsg);
    void (*unload_func) (NGPlugin *plugin_data);
    ngboolean (*can_unload_func) (NGPlugin *plugin_data, char **reason);
    void (*destroy_func) (NGPlugin *plugin_data);
    int (*get_version_func) (void);
};

typedef struct _ng_plugin_info {
    ngboolean is_loaded;

    char name[128];
    char version[64];
    char author[128];
    char url[256];
    char description[4096];

    NGList *dependencies;

    ngboolean initialisation_error;
    char initialisation_errmsg[4096];
} NGPluginInfo;

gboolean plugins_initialize(char **errmsg);
gboolean plugins_load(Configuration *configuration, char **errmsg);
void plugins_destroy(void);
gboolean load_plugin_by_name(const char *name, char **errmsg);
void destroy_plugin_by_name(const char *name);

/**
 * Create a new (internal) function which can be called from other plugins
 *
 * @param function_name The name of the function. Must be unique
 * @param impl          Pointer to the implementation of this function
 * @param marshaller    The function to translate arrays of parameter values to proper parameters
 * @param return_type   The return type of the function
 * @param num_params    The number of arguments which this function expects
 * @param ...           A list of parameter types
 */
ngboolean ng_plugin_register_internal_function(const char *function_name, NGPluginFunction impl, GSignalCMarshaller marshaller, GType return_type, int num_params, ...);

/**
 * Create a new (internal) event on which other plugins can register to
 *
 * @param event_name    The name of the event. Must be unique
 * @param marshaller    The function to translate arrays of parameter values to proper parameters
 * @param return_type   The return type of the event
 * @param num_params    The number of arguments which this event expects
 * @param ...           A list of parameter types
 */
ngboolean ng_plugin_create_internal_event(const char *event_name, GSignalCMarshaller marshaller, GType return_type, int num_params, ...);

/**
 * Emit an event
 *
 * @param event_name    The name of the event which needs to be emit
 * @param ...           Variable number of arguments to pass on to the event handlers
 * @return              TRUE if the event was successfully emit. FALSE if the event_name isn't registered yet
 */
ngboolean ng_plugin_emit_internal_event(const char *event_name, ...);

NGList *plugins_get_avail_plugins (void);
void plugins_free_avail_plugins (NGList *list);
ngboolean plugins_get_plugin_info (const char *plugin_name, NNTPGrabPluginInfo *plugin_info);
ngboolean plugins_load_plugin_by_name (const char *plugin_name, char **errmsg);
ngboolean plugins_unload_plugin_by_name (const char *plugin_name, char **errmsg);
ngboolean plugins_set_persistent (const char *plugin_name, ngboolean persistent);

#endif /* _NNTPGRAB_PLUGINS_H_ */
