/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Parts of this file are taken from ProFTP's throttle.c */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>
#ifdef WIN32
#include <windows.h>
#endif
#include "nntpgrab.h"
#include "nntpgrab_plugin.h"

/* This allows a throttling process to be killed by the admin */
#ifndef WIN32
static void
xfer_rate_sigmask(int block)
{
    static sigset_t sig_set;

    if (block) {
        sigemptyset(&sig_set);

        sigaddset(&sig_set, SIGCHLD);
        sigaddset(&sig_set, SIGUSR1);
        sigaddset(&sig_set, SIGINT);
        sigaddset(&sig_set, SIGQUIT);
#ifdef SIGIO
        sigaddset(&sig_set, SIGIO);
#endif /* SIGIO */
#ifdef SIGBUS
        sigaddset(&sig_set, SIGBUS);
#endif /* SIGBUS */
        sigaddset(&sig_set, SIGHUP);

        while (sigprocmask(SIG_BLOCK, &sig_set, NULL) < 0) {
            if (errno == EINTR) {
                continue;
            }

            break;
        }
    } else {
        while (sigprocmask(SIG_UNBLOCK, &sig_set, NULL) < 0) {
            if (errno == EINTR) {
                continue;
            }

            break;
        }
    }
}
#endif

/* Returns the difference, in milliseconds, between the given timeval and now */
static long
xfer_rate_since(struct timeval *then)
{
    struct timeval now;
    gettimeofday(&now, NULL);

    return (((now.tv_sec - then->tv_sec) * 1000L) + ((now.tv_usec - then->tv_usec) / 1000L));
}

void
throttle_pause(struct timeval start_time, off_t xferlen, int max_bandwidth)
{
    long double xfer_rate_bps;
    long ideal = 0, elapsed = 0;

    /* Calculate the time interval since the transfer of data started. */
    elapsed = xfer_rate_since(&start_time);
    xfer_rate_bps = max_bandwidth * 1024.0;

    ideal = (xferlen * 1000L) / xfer_rate_bps;

    if (ideal > elapsed) {
        struct timeval tv;

        /* Setup for the select.  We use select() instead of usleep() because it
         * seems to be far more portable across platforms.
         *
         * ideal and elapsed are in milleconds, but tv_usec will be microseconds,
         * so be sure to convert properly.
         */
        tv.tv_usec = (ideal - elapsed) * 1000;
        tv.tv_sec = tv.tv_usec / 1000000L;
        tv.tv_usec = tv.tv_usec % 1000000L;

        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_DEBUG, "transferring too fast, delaying %ld sec%s, %ld usecs",
            (long int) tv.tv_sec, tv.tv_sec == 1 ? "" : "s", (long int) tv.tv_usec);

#ifdef WIN32
        /* Stupid Win32 doesn't support the select() function with no sockets attached.. */
        Sleep(ideal - elapsed);
#else
        /* No interruptions, please... */
        xfer_rate_sigmask(TRUE);

        if (select(0, NULL, NULL, NULL, &tv) < 0) {
            ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_WARNING, "warning: unable to throttle bandwidth: %s", strerror(errno));
        }

        xfer_rate_sigmask(FALSE);
#endif
    }

    return;
}
