/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <errno.h>
#include <glib/gstdio.h>
#include <glib/gi18n.h>
#include "nntpgrab_plugin.h"
#include "config.h"

static ngboolean open_logfile(NGPlugin *plugin_data, char **errmsg);
static void close_logfile(NGPlugin *plugin_data);
static void on_config_changed(NGPlugin *plugin_data, gpointer data);
static void log_message_cb(NGPlugin *plugin_data, const char *component, NGLogLevel log_level, const char *message, gpointer data);

void
nntpgrab_plugin_initialize(NGPlugin *plugin_data)
{
    ng_plugin_set_name(plugin_data, "Logger");
    ng_plugin_set_version(plugin_data, PACKAGE_VERSION);
    ng_plugin_set_author(plugin_data, "Erik van Pienbroek");
    ng_plugin_set_url(plugin_data, "http://www.nntpgrab.nl");
    ng_plugin_set_description(plugin_data, "Save all log messages to disk (~/.config/NNTPGrab/nntpgrab.log)");

    ng_plugin_set_required_event(plugin_data, "log_message");
    ng_plugin_set_required_event(plugin_data, "config_changed");
}

ngboolean
nntpgrab_plugin_load(NGPlugin *plugin_data, char **errmsg)
{
    ConfigOpts opts = plugin_data->core_funcs.config_get_opts();

    ng_plugin_connect_event(plugin_data, "config_changed", NG_PLUGIN_FUNCTION(on_config_changed), NULL);

    if (opts.enable_logger) {
        return open_logfile(plugin_data, errmsg);
    } else {
        return TRUE;
    }
}

static ngboolean
open_logfile(NGPlugin *plugin_data, char **errmsg)
{
    char *filename = g_build_filename(g_get_user_config_dir(), "NNTPGrab", "nntpgrab.log", NULL);
    plugin_data->priv = g_fopen(filename, "w");
    if (!plugin_data->priv) {
        *errmsg = g_strdup_printf(_("Unable to open logfile '%s': %s"), filename, strerror(errno));
        g_free(filename);
        return FALSE;
    }
    g_free(filename);

    ng_plugin_connect_event(plugin_data, "log_message", NG_PLUGIN_FUNCTION(log_message_cb), NULL);

    return TRUE;
}

ngboolean
nntpgrab_plugin_can_unload(NGPlugin *plugin_data, char **reason)
{
    return TRUE;
}

void
nntpgrab_plugin_unload(NGPlugin *plugin_data)
{
    ConfigOpts opts = plugin_data->core_funcs.config_get_opts();

    if (opts.enable_logger) {
        close_logfile(plugin_data);
    }
}

static void
close_logfile(NGPlugin *plugin_data)
{
    g_return_if_fail(plugin_data != NULL);
    g_return_if_fail(plugin_data->priv != NULL);

    fclose(plugin_data->priv);
    plugin_data->priv = NULL;

    ng_plugin_disconnect_event_by_func(plugin_data, NG_PLUGIN_FUNCTION(log_message_cb), NULL);
}

void
nntpgrab_plugin_destroy(NGPlugin *plugin_data)
{
}

int
nntpgrab_plugin_get_version(void)
{
    return NNTPGRAB_PLUGIN_API_VERSION;
}

static void
on_config_changed(NGPlugin *plugin_data, gpointer data)
{
    ConfigOpts opts = plugin_data->core_funcs.config_get_opts();

    /* Do we need to start the logger? */
    if (opts.enable_logger && plugin_data->priv == NULL) {
        char *errmsg = NULL;

        if (!open_logfile(plugin_data, &errmsg)) {
            ng_plugin_emit_log_msg(plugin_data, NG_LOG_LEVEL_WARNING, _("Unable to start logger: %s"), errmsg);
            g_free(errmsg);
        }

        return;
    }

    /* Do we need to stop the logger? */
    if (!opts.enable_logger && plugin_data->priv != NULL) {
        close_logfile(plugin_data);
    }
}

static void
log_message_cb(NGPlugin *plugin_data, const char *component, NGLogLevel log_level, const char *message, gpointer data)
{
    char stamp_str[64];
    char *log_level_str;
    struct tm *tm;
    time_t stamp;

    g_return_if_fail(plugin_data->priv != NULL);

    memset(&stamp_str, 0, sizeof(stamp_str));
    stamp = time(NULL);
    tm = localtime(&stamp);
    strftime(stamp_str, sizeof(stamp_str) - 1, "%Y-%m-%d %H:%M:%S", tm);

    switch (log_level) {
        case NG_LOG_LEVEL_INFO:
            log_level_str = "INFO";
            break;

        case NG_LOG_LEVEL_WARNING:
            log_level_str = "WARNING";
            break;

        case NG_LOG_LEVEL_ERROR:
            log_level_str = "ERROR";
            break;

        case NG_LOG_LEVEL_FATAL:
            log_level_str = "FATAL";
            break;

        case NG_LOG_LEVEL_DEBUG:
            log_level_str = "DEBUG";
            break;

        case NG_LOG_LEVEL_ALL:
        default:
            /* Shouldn't happen */
            log_level_str = "UNKNOWN";
            break;
    }

    fprintf(plugin_data->priv, "%s\t%s\t%s\t%s\n", stamp_str, log_level_str, component, message);
    fflush(plugin_data->priv);
}
